import importlib
import multiprocessing as mp
from functools import partial

import cosmoprimo
import mpmath
import numpy as np
from scipy import integrate
from scipy.signal import savgol_filter
from scipy.special import spherical_jn

from flip.covariance import cov_utils
from flip.utils import create_log

log = create_log()
_avail_models = [
    "adamsblake17plane",
    "adamsblake17",
    "adamsblake20",
    "lai22",
    "carreres23",
    "ravouxcarreres",
    "ravouxnoanchor25",
    "rcrk24",
]
_avail_regularization_option = [
    None,
    "mpmath",
    "savgol",
    "lowk_asymptote",
]


def correlation_integration(ell, r, k, integrand):
    """
    The correlation_integration function is used to calculate the correlation function for a given multipole.
    It does this by integrating over k, which is the magnitude of the wavevector. The integration is performed using
    the Simpson's rule (integrate.simps). This function takes in four arguments: l, r, k and integrand. l represents
    the multipole that we are calculating for; it can be 0 (monopole), 2 (quadrupole) or 4 (hexadecapole). r represents
    the radial distance at which we want to evaluate our correlation function at; it can be any value between 0 and 200 M

    Args:
        l: Determine the order of the spherical bessel function
        r: Calculate the spherical bessel function
        k: Calculate the spherical bessel function
        integrand: Calculate the integrand of the correlation function

    Returns:
        ndarray: Correlation function $xi_\ell(r)$ via Simpson integration.

    """
    kr = np.outer(k, r)
    integrand = (
        (-1) ** (ell // 2)
        * (k**2 / (2 * np.pi**2))
        * integrand
        * spherical_jn(ell, kr).T
    )
    return (-1) ** (ell % 2) * integrate.simpson(integrand, x=k)


def correlation_hankel(ell, r, k, integrand, hankel_overhead_coefficient=2, kmin=None):
    """
    The correlation_hankel function is a wrapper for the cosmoprimo.fftlog.PowerToCorrelation function,
    which computes the correlation function from power spectrum using FFTLog (Hamilton 2000).
    The PowerToCorrelation class takes in an array of k values and an array of P(k) values, and returns
    an array of r values and an array of xi(r) values. The PowerToCorrelation class has two methods: set_fft_engine()
    and __call__(). The set_fft_engine() method sets which fft engine to use; it can be either &quot;n

    Args:
        l: Determine the parity of the integrand, and therefore whether to add a 1j term
        r: Set the radius of the sphere
        k: Set the kmin and kmax values in the powertocorrelation function
        integrand: Pass the integrand to the correlation_hankel function
        hankel_overhead_coefficient: Determine the minimum r value for which to use the hankel transform

    Returns:
        ndarray: Correlation function $xi_\ell(r)$ combining Hankel and direct integration.

    Note:
        If l is odd, count a 1j term in the integrand, without the need for adding it
    """
    Hankel = cosmoprimo.fftlog.PowerToCorrelation(k, ell=ell, q=0, complex=False)
    Hankel.set_fft_engine("numpy")
    r_hankel, xi_hankel = Hankel(integrand)
    mask = r < np.min(r_hankel) * hankel_overhead_coefficient
    if np.any(r > np.max(r_hankel)):
        raise ValueError(
            "Min pw spectrum k is too high, please take a lower one. Use kmin parameter to lower bound integration."
        )
    output = np.empty_like(r)
    output[mask] = correlation_integration(ell, r[mask], k, integrand)
    output[~mask] = (-1) ** (ell % 2) * np.interp(r[~mask], r_hankel, xi_hankel)

    # Regularization
    if kmin is not None:
        kreg = np.geomspace(np.min(k), kmin, int(len(k) / 20))
        output -= correlation_integration(ell, r, kreg, np.interp(kreg, k, integrand))
    return output


def coefficient_hankel(
    model_name,
    covariance_type,
    term_index,
    lmax,
    wavenumber,
    power_spectrum,
    coord,
    additional_parameters_values=None,
    **kwargs,
):
    """
    The coefficient_hankel function computes the covariance between two terms of a given model.
    The function takes as input:
        - The name of the model (e.g., adamsblake20, lai22, carreres23 or ravouxcarreres)
        - The type of term (i.e., either 'delta' or 'v')
        - The index i for the first term in question (i = 1 corresponds to delta_0 and v_0; i = 2 corresponds to delta_2 and v_2; etc.)
        - lmax: maximum multipole moment

    Args:
        model_name: Call the dictionary_subterms function from the flip_terms file
        type: Determine which subterms to use
        term_index: Select the term in the dictionary_subterms
        lmax: Determine the number of terms in the sum
        wavenumber: Calculate the hankel transform
        power_spectrum: Calculate the correlation function
        coord: Pass the coordinates of the point where we want to evaluate
        additional_parameters_values: Pass the values of additional parameters to the functions that
        : Define the model name

    Returns:
        ndarray: Covariance contribution for term `term_index`.

    """
    flip_terms = importlib.import_module(
        f"flip.covariance.analytical.{model_name}.flip_terms", package=__package__
    )

    cov_ab_i = 0
    flip_terms.set_backend("numpy")
    dictionary_subterms = flip_terms.dictionary_subterms
    regularize_M_terms = flip_terms.regularize_M_terms

    for ell in range(lmax + 1):
        number_terms = dictionary_subterms[f"{covariance_type}_{term_index}_{ell}"]
        for j in range(number_terms):
            M_ab_i_l_j = getattr(
                flip_terms, f"M_{covariance_type}_{term_index}_{ell}_{j}"
            )
            M_ab_i_l_j_evaluated = regularize_M(
                M_ab_i_l_j,
                wavenumber,
                regularize_M_terms,
                covariance_type,
                flip_terms,
                additional_parameters_values,
            )

            N_ab_i_l_j = getattr(
                flip_terms, f"N_{covariance_type}_{term_index}_{ell}_{j}"
            )(coord[1], coord[2])
            hankel_ab_i_l_j = correlation_hankel(
                ell,
                coord[0],
                wavenumber,
                M_ab_i_l_j_evaluated * power_spectrum,
                **kwargs,
            )
            cov_ab_i += N_ab_i_l_j * hankel_ab_i_l_j
    return cov_ab_i


def coefficient_trapz(
    model_name,
    covariance_type,
    term_index,
    lmax,
    wavenumber,
    power_spectrum,
    coord,
    additional_parameters_values=None,
    **kwargs,
):
    """
    The coefficient_trapz function computes the covariance matrix element for a given term.

    Args:
        model_name: Call the dictionary_subterms function from the flip_terms file
        type: Distinguish between the different terms in the covariance matrix
        term_index: Specify which term of the model is being used
        lmax: Specify the maximum order of spherical harmonics
        wavenumber: Calculate the integrand
        power_spectrum: Multiply the integrand
        coord: Pass in the coordinates of the point at which we want to evaluate
        additional_parameters_values: Pass the values of the additional parameters
        : Determine the model to be used

    Returns:
        ndarray: Covariance contribution for term `term_index`.

    """
    cov_ab_i = 0
    flip_terms = importlib.import_module(
        f"flip.covariance.analytical.{model_name}.flip_terms", package=__package__
    )
    flip_terms.set_backend("numpy")
    dictionary_subterms = flip_terms.dictionary_subterms
    regularize_M_terms = flip_terms.regularize_M_terms

    for ell in range(lmax + 1):
        number_terms = dictionary_subterms[f"{covariance_type}_{term_index}_{ell}"]
        for j in range(number_terms):
            M_ab_i_l_j = getattr(
                flip_terms, f"M_{covariance_type}_{term_index}_{ell}_{j}"
            )
            M_ab_i_l_j_evaluated = regularize_M(
                M_ab_i_l_j,
                wavenumber,
                regularize_M_terms,
                covariance_type,
                flip_terms,
                additional_parameters_values,
            )
            N_ab_i_l_j = getattr(
                flip_terms, f"N_{covariance_type}_{term_index}_{ell}_{j}"
            )(coord[1], coord[2])
            kr = np.outer(wavenumber, coord[0])
            integrand = (
                (-1) ** (ell // 2)
                * (wavenumber**2 / (2 * np.pi**2))
                * M_ab_i_l_j_evaluated
                * power_spectrum
                * spherical_jn(ell, kr).T
            )
            hankel_ab_i_l_j = (-1) ** (ell % 2) * np.trapezoid(integrand, x=wavenumber)
            cov_ab_i += N_ab_i_l_j * hankel_ab_i_l_j
    return cov_ab_i


def regularize_M(
    M_function,
    wavenumber,
    regularize_M_terms,
    covariance_type,
    flip_terms,
    additional_parameters_values,
    savgol_window=50,
    savgol_polynomial=3,
    lowk_unstable_threshold=0.1,
    lowk_unstable_mean_filtering=10,
    mpmmath_decimal_precision=50,
):
    """Evaluate and optionally regularize M(k) term.

    Applies one of the supported regularizations to stabilize numerical
    behavior (mpmath high-precision, Savitzky–Golay smoothing, or low-k
    asymptote detection).

    Args:
        M_function (callable): Function returning M(k) given additional parameters.
        wavenumber (ndarray): k samples.
        regularize_M_terms (dict|None): Per type regularization option.
        covariance_type (str): `"gg"`, `"gv"`, or `"vv"`.
        flip_terms (module): Terms module to switch backend when needed.
        additional_parameters_values (dict|tuple): Extra parameters for M.
        savgol_window (int): Window size for Savitzky–Golay.
        savgol_polynomial (int): Polynomial order for Savitzky–Golay.
        lowk_unstable_threshold (float): Threshold for low-k asymptote detection.
        lowk_unstable_mean_filtering (int): Window for mean filtering indices.
        mpmmath_decimal_precision (int): Decimal precision for mpmath.

    Returns:
        ndarray: Evaluated M(k) after regularization.
    """
    if regularize_M_terms is None:
        return M_function(*additional_parameters_values)(wavenumber)
    else:
        regularization_option = regularize_M_terms[covariance_type]

        if regularization_option is None:
            return M_function(*additional_parameters_values)(wavenumber)

        elif regularization_option == "mpmath":
            flip_terms.set_backend("mpmath")
            mpmath.mp.dps = mpmmath_decimal_precision
            wavenumber_mpmath = wavenumber * mpmath.mpf(1)
            additional_parameters_values_mpf = tuple(
                [mpmath.mpf(par) for par in additional_parameters_values]
            )
            M_function_evaluated = np.array(
                np.frompyfunc(M_function(*additional_parameters_values_mpf), 1, 1)(
                    wavenumber_mpmath
                ).tolist(),
                dtype=float,
            )
            flip_terms.set_backend("numpy")

        elif regularization_option == "savgol":
            M_function_evaluated = M_function(*additional_parameters_values)(wavenumber)
            M_function_evaluated = savgol_filter(
                M_function_evaluated,
                savgol_window,
                savgol_polynomial,
            )

        elif regularization_option == "lowk_asymptote":
            # The low k region presents numerical instabilities for density models.
            # All the M density function should present and asymptotic behaviour at low k.
            # This method detect low k asymptote and force it for all M functions.
            M_function_evaluated = M_function(*additional_parameters_values)(wavenumber)
            diff = np.diff(M_function_evaluated, append=[M_function_evaluated[-1]])
            mask_asymptote = np.abs(diff) < lowk_unstable_threshold * np.mean(
                np.abs(diff[wavenumber > wavenumber[len(wavenumber) // 2]])
            )
            mask_asymptote &= wavenumber < wavenumber[3 * len(wavenumber) // 4]

            if len(mask_asymptote[mask_asymptote]) > 0:
                index_mask_asymptote = np.argwhere(mask_asymptote)[:, 0]
                diff_index = np.diff(
                    index_mask_asymptote, prepend=[index_mask_asymptote[0]]
                )
                mean_diff_index = np.convolve(
                    diff_index,
                    np.ones(lowk_unstable_mean_filtering)
                    / lowk_unstable_mean_filtering,
                    mode="same",
                )
                mask_best_value_asymptote = np.abs(mean_diff_index - 1.0) < 10**-5
                if len(mask_best_value_asymptote[mask_best_value_asymptote]) == 0:
                    index_asymptote = index_mask_asymptote[-1]
                else:
                    index_asymptote = index_mask_asymptote[mask_best_value_asymptote][0]
                mask_unstable_region = wavenumber < wavenumber[index_asymptote]
                M_function_evaluated[mask_unstable_region] = M_function_evaluated[
                    index_asymptote
                ]
        else:
            raise ValueError(
                f"regularization option {regularization_option} is not available"
                f"Please choose in: {_avail_regularization_option}"
            )

    return M_function_evaluated


def compute_coordinates(
    covariance_type,
    coordinates_density=None,
    coordinates_velocity=None,
    size_batch=10_000,
    los_definition="bisector",
):
    """
    The compute_coordinates function computes the spherical coordinates of all pairs of objects in a given catalog.

    Args:
        covariance_type: Determine whether the covariance is a cross-covariance or an auto-covariance
        coordinates_density: Store the coordinates of the density field
        coordinates_velocity: Pass the coordinates of the velocity field
        size_batch: Control the size of the batches
        : Compute the angle separation between two objects

    Returns:
        list: Per-batch arrays `[r, theta, phi]`.
    """
    if covariance_type == "gg":
        ra = coordinates_density[0]
        dec = coordinates_density[1]
        comoving_distance = coordinates_density[2]
        number_objects = len(ra)
    elif covariance_type == "vv":
        ra = coordinates_velocity[0]
        dec = coordinates_velocity[1]
        comoving_distance = coordinates_velocity[2]
        number_objects = len(ra)
    elif covariance_type == "gv":
        ra_g = coordinates_density[0]
        dec_g = coordinates_density[1]
        comoving_distance_g = coordinates_density[2]
        ra_v = coordinates_velocity[0]
        dec_v = coordinates_velocity[1]
        comoving_distance_v = coordinates_velocity[2]
        number_objects_g = len(ra_g)
        number_objects_v = len(ra_v)

    if (covariance_type == "gg") | (covariance_type == "vv"):
        cross = False
        n_task = int((number_objects * (number_objects + 1)) / 2) - number_objects
    else:
        cross = True
        n_task = int(number_objects_g * number_objects_v)

    parameters = []
    for n in range(0, n_task, size_batch):
        batches = np.arange(n, np.min((n + size_batch, n_task)))
        if cross:
            i_list, j_list = cov_utils.compute_i_j_cross_matrix(
                number_objects_v, batches
            )
            ra_i, dec_i, r_i = (
                ra_g[i_list],
                dec_g[i_list],
                comoving_distance_g[i_list],
            )
            ra_j, dec_j, r_j = (
                ra_v[j_list],
                dec_v[j_list],
                comoving_distance_v[j_list],
            )
        else:
            i_list, j_list = cov_utils.compute_i_j(number_objects, batches)
            ra_i, dec_i, r_i = (
                ra[i_list],
                dec[i_list],
                comoving_distance[i_list],
            )
            ra_j, dec_j, r_j = (
                ra[j_list],
                dec[j_list],
                comoving_distance[j_list],
            )
        r, theta, phi = cov_utils.angle_separation(
            ra_i, ra_j, dec_i, dec_j, r_i, r_j, los_definition=los_definition
        )
        parameters.append([r, theta, phi])
    return parameters


def compute_coeficient(
    parameters,
    model_name,
    covariance_type,
    power_spectrum_list,
    additional_parameters_values=None,
    number_worker=8,
    hankel=True,
    kmin=None,
):
    """
    The compute_coeficient function computes the covariance matrix for a given model.

    Args:
        parameters: Compute the covariance matrix for each parameter
        model_name: Select the model to be used
        covariance_type: Select the type of covariance we want to compute
        power_spectrum_list: Compute the power spectrum of each term in the covariance matrix
        additional_parameters_values: Pass the values of additional parameters, such as the redshift
        number_worker: Specify the number of cores to use for multiprocessing
        hankel: Choose between the trapezoidal rule and the hankel transform
        : Define the number of threads used to compute the covariance matrix

    Returns:
        ndarray: Stacked covariance arrays per term index.
    """
    if additional_parameters_values is None:
        additional_parameters_values = {}
    if hankel:
        coefficient = coefficient_hankel
    else:
        coefficient = coefficient_trapz

    flip_terms = importlib.import_module(
        f"flip.covariance.analytical.{model_name}.flip_terms", package=__package__
    )
    term_index_list = getattr(flip_terms, "dictionary_terms")[covariance_type]
    lmax_list = getattr(flip_terms, "dictionary_lmax")[covariance_type]
    multi_index_model = getattr(flip_terms, "multi_index_model")

    function_covariance_dict = {}
    for i, index in enumerate(term_index_list):
        if multi_index_model:
            index_power_spectrum = int(index[0])
        else:
            index_power_spectrum = i

        function_covariance_dict[f"func_{index}"] = partial(
            coefficient,
            model_name,
            covariance_type,
            index,
            lmax_list[i],
            power_spectrum_list[index_power_spectrum][0],
            power_spectrum_list[index_power_spectrum][1],
            additional_parameters_values=additional_parameters_values,
            kmin=kmin,
        )
    if number_worker == 1:
        for i, index in enumerate(term_index_list):
            function_covariance_dict[f"cov_{index}"] = np.concatenate(
                [
                    function_covariance_dict[f"func_{index}"](param)
                    for param in parameters
                ]
            )
    else:
        with mp.Pool(number_worker) as pool:
            for i, index in enumerate(term_index_list):
                function_covariance_dict[f"map_async_{index}"] = pool.map_async(
                    function_covariance_dict[f"func_{index}"],
                    parameters,
                )
            for i, index in enumerate(term_index_list):
                function_covariance_dict[f"cov_{index}"] = np.concatenate(
                    function_covariance_dict[f"map_async_{index}"].get()
                )

    # In the case of autocorrelation, add the theoretical variance.
    if covariance_type[0] == covariance_type[1]:
        for i, index in enumerate(term_index_list):
            if multi_index_model:
                index_power_spectrum = int(index[0])
            else:
                index_power_spectrum = i
            variance_t = coefficient(
                model_name,
                covariance_type,
                index,
                lmax_list[i],
                power_spectrum_list[index_power_spectrum][0],
                power_spectrum_list[index_power_spectrum][1],
                np.zeros((5, 1)),
                additional_parameters_values=additional_parameters_values,
            )[0]

            function_covariance_dict[f"cov_{index}"] = np.insert(
                function_covariance_dict[f"cov_{index}"],
                0,
                variance_t,
            )

    return np.array(
        [
            function_covariance_dict[f"cov_{index}"]
            for _, index in enumerate(term_index_list)
        ]
    )


def compute_cov(
    model_name,
    covariance_type,
    power_spectrum_list,
    coordinates_density=None,
    coordinates_velocity=None,
    additional_parameters_values=None,
    size_batch=10_000,
    number_worker=8,
    hankel=True,
    los_definition="bisector",
    kmin=None,
):
    """
    The compute_cov function computes the covariance matrix for a given model.

    Args:
        model_name: Choose the model to be used
        covariance_type: Determine the type of covariance to be computed
        power_spectrum_list: Compute the power spectrum of the density field
        coordinates_density: Set the range of values for the density coordinates
        coordinates_velocity: Compute the velocity covariance
        additional_parameters_values: Pass the values of the additional parameters to be used in the model
        size_batch: Split the computation of the covariance matrix in smaller matrices
        number_worker: Specify the number of cores to be used
        hankel: Choose between the hankel transform or the fourier transform
        : Compute the covariance matrix for a given model

    Returns:
        ndarray: Covariance arrays per term index.

    """
    if model_name not in _avail_models:
        log.add(
            f"Model {model_name} not available."
            f"Please choose between: {_avail_models}"
        )

    parameters = compute_coordinates(
        covariance_type,
        coordinates_density=coordinates_density,
        coordinates_velocity=coordinates_velocity,
        size_batch=size_batch,
        los_definition=los_definition,
    )
    covariance = compute_coeficient(
        parameters,
        model_name,
        covariance_type,
        power_spectrum_list,
        additional_parameters_values=additional_parameters_values,
        number_worker=number_worker,
        hankel=hankel,
        kmin=kmin,
    )

    return covariance


def generate_covariance(
    model_name,
    model_kind,
    power_spectrum_dict,
    coordinates_velocity=None,
    coordinates_density=None,
    additional_parameters_values=None,
    size_batch=10_000,
    number_worker=8,
    hankel=True,
    los_definition="bisector",
    kmin=None,
):
    """
    The generate_flip function computes the covariance matrix for a given model.

    Args:
        model_name: Select the model to use
        model_kind: Determine the type of model to generate
        power_spectrum_dict: Store the power spectra of the different fields
        coordinates_velocity: Specify the coordinates of the velocity field
        coordinates_density: Specify the coordinates of the density field
        additional_parameters_values: Pass the values of the additional parameters to be used in the computation of covariance matrices
        size_batch: Split the computation of the covariance matrix into smaller batches
        number_worker: Specify the number of cores to use for computing the covariance matrix
        hankel: Decide whether to use the hankel transform or not
        : Define the number of workers to use for the computation

    Returns:
        tuple: `(covariance_dict, number_densities, number_velocities)`.

    """
    cov_utils.check_generator_need(
        model_kind,
        coordinates_density,
        coordinates_velocity,
    )
    covariance_dict = {}

    if model_kind in ["density", "full", "density_velocity"]:
        covariance_dict["gg"] = compute_cov(
            model_name,
            "gg",
            power_spectrum_dict["gg"],
            coordinates_density=coordinates_density,
            coordinates_velocity=coordinates_velocity,
            additional_parameters_values=additional_parameters_values,
            size_batch=size_batch,
            number_worker=number_worker,
            hankel=hankel,
            los_definition=los_definition,
            kmin=kmin,
        )
        number_densities = len(coordinates_density[0])
    else:
        number_densities = None

    if model_kind in ["velocity", "full", "density_velocity"]:
        covariance_dict["vv"] = compute_cov(
            model_name,
            "vv",
            power_spectrum_dict["vv"],
            coordinates_density=coordinates_density,
            coordinates_velocity=coordinates_velocity,
            additional_parameters_values=additional_parameters_values,
            size_batch=size_batch,
            number_worker=number_worker,
            hankel=hankel,
            los_definition=los_definition,
            kmin=kmin,
        )
        number_velocities = len(coordinates_velocity[0])
    else:
        number_velocities = None

    if model_kind == "full":
        covariance_dict["gv"] = compute_cov(
            model_name,
            "gv",
            power_spectrum_dict["gv"],
            coordinates_density=coordinates_density,
            coordinates_velocity=coordinates_velocity,
            additional_parameters_values=additional_parameters_values,
            size_batch=size_batch,
            number_worker=number_worker,
            hankel=hankel,
            los_definition=los_definition,
            kmin=kmin,
        )

    return covariance_dict, number_densities, number_velocities



import os

import numpy as np
import pandas as pd
from flip.covariance import covariance, fitter
from pkg_resources import resource_filename

from flip import data_vector, utils

flip_base = resource_filename("flip", ".")
data_path = os.path.join(flip_base, "data")

### Load data
sn_data = pd.read_parquet(os.path.join(data_path, "data_velocity.parquet"))

coordinates_velocity = np.array([sn_data["ra"], sn_data["dec"], sn_data["rcom_zobs"]])

data_velocity = sn_data.to_dict("list")
for key in data_velocity.keys():
    data_velocity[key] = np.array(data_velocity[key])
data_velocity["velocity"] = data_velocity.pop("vpec")
data_velocity["velocity_error"] = np.zeros_like(data_velocity["velocity"])

data_velocity_object = data_vector.DirectVel(data_velocity)


ktt, ptt = np.loadtxt(os.path.join(data_path, "power_spectrum_tt.txt"))
kmt, pmt = np.loadtxt(os.path.join(data_path, "power_spectrum_mt.txt"))
kmm, pmm = np.loadtxt(os.path.join(data_path, "power_spectrum_mm.txt"))

sigmau_fiducial = 15

power_spectrum_dict = {"vv": [[ktt, ptt * utils.Du(ktt, sigmau_fiducial) ** 2]]}

### Compute covariance
sigmau_list = np.linspace(10.0, 20.0, 10)
covariance_list = []
size_batch = 10_000
number_worker = 16


for sigu in sigmau_list:
    power_spectrum_dict = {"vv": [[ktt, ptt * utils.Du(ktt, sigu) ** 2]]}

    covariance_list.append(
        covariance.CovMatrix.init_from_flip(
            "carreres23",
            "velocity",
            power_spectrum_dict,
            coordinates_velocity=coordinates_velocity,
            size_batch=size_batch,
            number_worker=number_worker,
        )
    )


### Load fitter
likelihood_type = "multivariate_gaussian_interp1d"
likelihood_properties = {"inversion_method": "cholesky_inverse"}

parameter_dict = {
    "fs8": {
        "value": 0.4,
        "limit_low": 0.0,
        "fixed": False,
    },
    "sigv": {
        "value": 200,
        "limit_low": 0.0,
        "fixed": False,
    },
    "sigu": {
        "value": 15.0,
        "limit_low": 13.0,
        "limit_up": 17.0,
        "fixed": False,
    },
}


minuit_fitter = fitter.FitMinuit.init_from_covariance(
    covariance_list,
    data_velocity_object,
    parameter_dict,
    likelihood_type=likelihood_type,
    likelihood_properties=likelihood_properties,
    interpolation_value_name="sigu",
    interpolation_value_range=sigmau_list,
)


### Fit
minuit_fitter.run()



_variant = [None, "baseline", "nobeta"]

_free_par = {
    "fs8": ["velocity@all", "density@nobeta"],
    "bs8": "density@all",
    "sigv": "velocity@all",
    "beta_f": "density@baseline",
}

_coordinate_keys = ["ra", "dec", "rcom_zobs"]



"""Init file of the flip.covariance package."""

from . import (
    analytical,
    contraction,
    cov_utils,
    emulators,
    fisher,
    fitter,
    generator,
    likelihood,
    symbolic,
)
from .covariance import CovMatrix



import numpy as np


def get_partial_derivative_coefficients(
    model_kind,
    parameter_values_dict,
    variant=None,
    covariance_prefactor_dict=None,
):
    if variant == "growth_index":
        partial_coefficients_dict = {
            "Omegam": {
                "vv": [
                    2
                    * parameter_values_dict["Omegam"]
                    ** (2 * parameter_values_dict["gamma"])
                    * parameter_values_dict["gamma"]
                    * parameter_values_dict["s8"] ** 2
                    / parameter_values_dict["Omegam"],
                ],
            },
            "gamma": {
                "vv": [
                    2
                    * parameter_values_dict["Omegam"]
                    ** (2 * parameter_values_dict["gamma"])
                    * parameter_values_dict["s8"] ** 2
                    * np.log(parameter_values_dict["Omegam"]),
                ],
            },
            "s8": {
                "vv": [
                    2
                    * parameter_values_dict["Omegam"]
                    ** (2 * parameter_values_dict["gamma"])
                    * parameter_values_dict["s8"],
                ],
            },
        }
    else:
        partial_coefficients_dict = {
            "fs8": {
                "vv": [
                    2 * parameter_values_dict["fs8"],
                ],
            },
        }
    return partial_coefficients_dict



import importlib
import os

import numpy as np

_available_engines = ["class_engine", "cosmoprimo_engine", "pyccl_engine"]
_available_power_spectrum_model = ["linearbel", "nonlinearbel", "linear"]
_available_power_spectrum_normalizaton = [
    "no_normalization",
    "growth_rate",
    "growth_amplitude",
]


def get_power_spectrum_suffix(
    redshift,
    minimal_wavenumber,
    maximal_wavenumber,
    number_points,
    log_space,
):
    """Build a filename suffix encoding spectrum sampling settings.

    Returns:
        str: Suffix including z, kmin, kmax, N, and log/lin tag.
    """
    return f"z{redshift}_kmin{minimal_wavenumber:.4f}_kmax{maximal_wavenumber:.4f}_N{number_points}{'_log' if log_space else '_lin'}"


def get_power_spectrum_name(
    power_spectrum_model,
    power_spectrum_type,
    suffix,
):
    """Construct a power spectrum filename for saving.

    Args:
        power_spectrum_model (str): Model name, e.g., `linearbel`.
        power_spectrum_type (str): One of `mm`, `mt`, `tt`.
        suffix (str): Sampling suffix from `get_power_spectrum_suffix`.

    Returns:
        str: Filename `power_spectrum_<model>_<type>_<suffix>.txt`.
    """
    return f"power_spectrum_{power_spectrum_model}_{power_spectrum_type}_{suffix}.txt"


def save_power_spectrum(
    wavenumber,
    power_spectrum,
    power_spectrum_model,
    power_spectrum_type,
    suffix,
    header,
    path,
):
    """Save a power spectrum to disk as a two-row text file.

    Args:
        wavenumber (ndarray): $k$ samples.
        power_spectrum (ndarray): Spectrum values.
        power_spectrum_model (str): Model name.
        power_spectrum_type (str): `mm`, `mt`, or `tt`.
        suffix (str): Sampling suffix.
        header (str): Header string with metadata.
        path (str): Directory where to save.
    """
    power_spectrum_name = get_power_spectrum_name(
        power_spectrum_model,
        power_spectrum_type,
        suffix,
    )

    np.savetxt(
        os.path.join(path, power_spectrum_name),
        [wavenumber, power_spectrum],
        header=header,
    )


def compute_power_spectra(
    power_spectrum_engine,
    power_spectrum_settings,
    redshift,
    minimal_wavenumber,
    maximal_wavenumber,
    number_points,
    logspace=True,
    normalization_power_spectrum="no_normalization",
    power_spectrum_non_linear_model=None,
    power_spectrum_model="linearbel",
    save_path=None,
):
    """Compute and optionally save MM/MT/TT power spectra.

    Args:
        power_spectrum_engine (str): Engine module, one of `_available_engines`.
        power_spectrum_settings (dict|object): Engine configuration.
        redshift (float): Target redshift.
        minimal_wavenumber (float): Minimum $k$ in h/Mpc.
        maximal_wavenumber (float): Maximum $k$ in h/Mpc.
        number_points (int): Number of $k$ samples.
        logspace (bool): Sample $k$ in log-space if True.
        normalization_power_spectrum (str): One of `_available_power_spectrum_normalizaton`.
        power_spectrum_non_linear_model (str|None): Non-linear engine flag for CLASS.
        power_spectrum_model (str): One of `_available_power_spectrum_model`.
        save_path (str|None): Directory to save spectra.

    Returns:
        tuple: `(k, P_mm, P_mt, P_tt, fiducial_dict)`.

    Raises:
        ValueError: If engine or model name is unsupported.
    """
    if power_spectrum_engine not in _available_engines:
        raise ValueError(
            f"The engine {power_spectrum_engine} is not available"
            f"PLease choose between {_available_engines}"
        )
    if power_spectrum_model not in _available_power_spectrum_model:
        raise ValueError(
            f"The power spectrum model {power_spectrum_model} is not available"
            f"PLease choose between {_available_power_spectrum_model}"
        )

    engine = importlib.import_module(f"flip.power_spectra.{power_spectrum_engine}")

    (
        wavenumber,
        power_spectrum_linear,
        power_spectrum_non_linear,
        fiducial,
    ) = engine.compute_power_spectrum(
        power_spectrum_settings,
        redshift,
        minimal_wavenumber,
        maximal_wavenumber,
        number_points,
        logspace=logspace,
        non_linear_model=power_spectrum_non_linear_model,
    )
    module = importlib.import_module("flip.power_spectra.models")
    model_function = getattr(module, f"get_{power_spectrum_model}_model")

    power_spectrum_mm, power_spectrum_mt, power_spectrum_tt = model_function(
        wavenumber,
        power_spectrum_linear,
        power_spectrum_non_linear=power_spectrum_non_linear,
        **fiducial,
    )

    if normalization_power_spectrum == "growth_rate":
        power_spectrum_mm = power_spectrum_mm
        power_spectrum_mt = power_spectrum_mt * (
            fiducial["fsigma_8"] / fiducial["sigma_8"]
        )
        power_spectrum_tt = (
            power_spectrum_tt * (fiducial["fsigma_8"] / fiducial["sigma_8"]) ** 2
        )
    elif normalization_power_spectrum == "growth_amplitude":
        power_spectrum_mm = power_spectrum_mm / fiducial["sigma_8"] ** 2
        power_spectrum_mt = power_spectrum_mt / fiducial["sigma_8"] ** 2
        power_spectrum_tt = power_spectrum_tt / fiducial["sigma_8"] ** 2
    elif normalization_power_spectrum == "no_normalization":
        pass
    else:
        raise ValueError(
            f"The normalization {normalization_power_spectrum} of the power spectrum is not available,"
            f"Please choose in {_available_power_spectrum_normalizaton}."
        )

    if save_path is not None:
        suffix = get_power_spectrum_suffix(
            redshift,
            minimal_wavenumber,
            maximal_wavenumber,
            number_points,
            logspace,
        )
        header = ""
        for key in fiducial.keys():
            header = header + f"fiducial {key} = {fiducial[key]} & "

        save_power_spectrum(
            wavenumber,
            power_spectrum_mm,
            power_spectrum_model,
            "mm",
            suffix,
            header,
            save_path,
        )
        save_power_spectrum(
            wavenumber,
            power_spectrum_mt,
            power_spectrum_model,
            "mt",
            suffix,
            header,
            save_path,
        )
        save_power_spectrum(
            wavenumber,
            power_spectrum_tt,
            power_spectrum_model,
            "tt",
            suffix,
            header,
            save_path,
        )

    return wavenumber, power_spectrum_mm, power_spectrum_mt, power_spectrum_tt, fiducial



def get_coefficients(
    parameter_values_dict,
    model_kind,
    variant=None,
    covariance_prefactor_dict=None,
):
    coefficients_dict = {}
    if model_kind in ["density", "full", "density_velocity"]:
        coefficients_dict["gg"] = [
            parameter_values_dict["bs8"] ** 2,
        ]
    if model_kind in ["full"]:
        coefficients_dict["gv"] = [
            parameter_values_dict["bs8"] * parameter_values_dict["fs8"],
        ]
    if model_kind in ["velocity", "full", "density_velocity"]:
        coefficients_dict["vv"] = [parameter_values_dict["fs8"] ** 2]
    return coefficients_dict


def get_diagonal_coefficients(parameter_values_dict, model_kind):
    coefficients_dict = {}
    if model_kind in ["density", "full", "density_velocity"]:
        coefficients_dict["gg"] = 0.0
    if model_kind in ["velocity", "full", "density_velocity"]:
        coefficients_dict["vv"] = parameter_values_dict["sigv"] ** 2
    return coefficients_dict



import numpy as np


def bel_coefficients(sigma_8):
    """Compute BEL model coefficients as a function of $\sigma_8$.

    Uses the parameterization from Bel et al. to construct coefficients
    for the non-linear damping terms.

    Args:
        sigma_8 (float): Amplitude of matter fluctuations in 8 Mpc/h spheres.

    Returns:
        tuple: `(a1, a2, a3, invkdelta, b)` coefficients.
    """
    a1 = -0.817 + 3.198 * sigma_8
    a2 = 0.877 - 4.191 * sigma_8
    a3 = -1.199 + 4.629 * sigma_8
    invkdelta = -0.017 + 1.496 * sigma_8**2
    b = 0.091 + 0.702 * sigma_8**2
    return a1, a2, a3, invkdelta, b


def get_bel_model(
    wavenumber,
    power_spectrum_linear,
    **kwargs,
):
    """Apply BEL damping to linear spectrum to get TT and MT forms.

    Args:
        wavenumber (ndarray): $k$ values in h/Mpc.
        power_spectrum_linear (ndarray): Linear matter power spectrum $P_{mm}^{lin}(k)$.
        **kwargs: Must include `sigma_8`.

    Returns:
        tuple: `(P_mt(k), P_tt(k))` BEL-damped spectra.
    """
    if "sigma_8" not in kwargs.keys():
        raise ValueError("Fiducial sigma_8 value is needed for nonlinearbel model")

    a1, a2, a3, invkdelta, b = bel_coefficients(kwargs["sigma_8"])

    power_spectrum_tt = power_spectrum_linear * np.exp(
        -wavenumber * (a1 + a2 * wavenumber + a3 * wavenumber**2)
    )

    power_spectrum_mt = power_spectrum_linear * np.exp(
        -invkdelta * wavenumber - b * wavenumber**6
    )

    return power_spectrum_mt, power_spectrum_tt


def get_nonlinearbel_model(
    wavenumber,
    power_spectrum_linear,
    **kwargs,
):
    """Return BEL-damped spectra using an external non-linear $P_{mm}$.

    Args:
        wavenumber (ndarray): $k$ values in h/Mpc.
        power_spectrum_linear (ndarray): Linear $P_{mm}^{lin}(k)$.
        **kwargs: Must include `power_spectrum_non_linear` and `sigma_8`.

    Returns:
        tuple: `(P_mm(k), P_mt(k), P_tt(k))` with $P_mm$ taken from engine.
    """
    if "power_spectrum_non_linear" not in kwargs.keys():
        raise ValueError("Non linear power spectrum is needed for nonlinearbel model")

    power_spectrum_mt, power_spectrum_tt = get_bel_model(
        wavenumber, power_spectrum_linear, **kwargs
    )

    power_spectrum_mt *= np.sqrt(
        kwargs["power_spectrum_non_linear"] / power_spectrum_linear
    )
    power_spectrum_mm = kwargs["power_spectrum_non_linear"]

    return power_spectrum_mm, power_spectrum_mt, power_spectrum_tt


def get_linearbel_model(
    wavenumber,
    power_spectrum_linear,
    **kwargs,
):
    """Return linear spectra with BEL damping applied to MT and TT.

    Args:
        wavenumber (ndarray): $k$ values in h/Mpc.
        power_spectrum_linear (ndarray): Linear $P_{mm}^{lin}(k)$.
        **kwargs: Must include `sigma_8`.

    Returns:
        tuple: `(P_mm, P_mt, P_tt)` with `P_mm = P_lin`.
    """
    power_spectrum_mt, power_spectrum_tt = get_bel_model(
        wavenumber, power_spectrum_linear, **kwargs
    )
    power_spectrum_mm = power_spectrum_linear

    return power_spectrum_mm, power_spectrum_mt, power_spectrum_tt


def get_linear_model(
    wavenumber,
    power_spectrum_linear,
    **kwargs,
):
    """Return purely linear spectra for MM, MT, and TT.

    Args:
        wavenumber (ndarray): $k$ values (unused).
        power_spectrum_linear (ndarray): Linear $P_{mm}^{lin}(k)$.

    Returns:
        tuple: `(P_mm, P_mt, P_tt)` all equal to `P_lin`.
    """

    return power_spectrum_linear, power_spectrum_linear, power_spectrum_linear



import importlib

from flip.utils import create_log

from .._config import __use_jax__

if __use_jax__:
    try:
        import jax.numpy as jnp

        jax_installed = True

    except ImportError:
        import numpy as jnp

        jax_installed = False
else:
    import numpy as jnp

    jax_installed = False

log = create_log()


def inverse_covariance_inverse(covariance):
    """Return the explicit inverse of a covariance matrix.

    Args:
        covariance (array-like): Positive-definite covariance matrix of shape `(N, N)`.

    Returns:
        numpy.ndarray: The matrix inverse `covariance^{-1}`.

    Notes:
        For Fisher computations, explicit inversion is acceptable on modest sizes,
        but Cholesky-based solves are usually more stable for ill-conditioned matrices.
    """
    return jnp.linalg.inv(covariance)


def inverse_covariance_cholesky(covariance):
    """Return the explicit inverse of a covariance matrix.

    Args:
        covariance (array-like): Positive-definite covariance matrix of shape `(N, N)`.

    Returns:
        numpy.ndarray: The matrix inverse `covariance^{-1}`.

    Notes:
        For Fisher computations, explicit inversion is acceptable on modest sizes,
        but Cholesky-based solves are usually more stable for ill-conditioned matrices.
    """
    c = jnp.linalg.inv(jnp.linalg.cholesky(covariance))
    inverse = jnp.dot(c.T, c)
    return inverse


def inverse_covariance_cholesky_inverse(covariance):
    """Return the explicit inverse of a covariance matrix.

    Args:
        covariance (array-like): Positive-definite covariance matrix of shape `(N, N)`.

    Returns:
        numpy.ndarray: The matrix inverse `covariance^{-1}`.

    Notes:
        For Fisher computations, explicit inversion is acceptable on modest sizes,
        but Cholesky-based solves are usually more stable for ill-conditioned matrices.
    """
    c = jnp.linalg.inv(jnp.linalg.cholesky(covariance))
    inverse = jnp.dot(c.T, c)
    return inverse


class FisherMatrix:
    """Compute Fisher information matrix from covariance derivatives.

    Builds the Fisher matrix $F_{ij} = \tfrac{1}{2}\,\mathrm{Tr}[C^{-1} \partial_i C\, C^{-1} \partial_j C]$
    using model-specific derivative coefficients and a provided inverse covariance.

    Attributes:
        covariance (CovMatrix): Covariance model describing blocks gg/gv/vv.
        inverse_covariance_sum (numpy.ndarray): Precomputed `C^{-1}` used in traces.
        parameter_values_dict (dict): Parameter values for computing coefficients.
        free_par (list[str]): Combined free parameters from covariance and data.
    """

    _default_fisher_properties = {
        "inversion_method": "inverse",
        "negative_log_likelihood": True,
    }

    def __init__(
        self,
        covariance=None,
        inverse_covariance_sum=None,
        data_free_par=None,
        parameter_values_dict=None,
    ):

        self.covariance = covariance
        self.inverse_covariance_sum = inverse_covariance_sum
        self.parameter_values_dict = parameter_values_dict
        self.free_par = self.covariance.free_par[:]
        if data_free_par is not None:
            self.free_par += data_free_par

    @classmethod
    def init_from_covariance(
        cls,
        covariance,
        data,
        parameter_values_dict,
        fisher_properties={},
        covariance_prefactor_dict=None,
    ):
        """Initialize a FisherMatrix instance from a covariance and data model.

        Ensures the covariance is in matrix form, prepares the `compute_covariance_sum`
        functions, computes the data variance and the full covariance sum, and inverts
        it according to the selected method.

        Args:
            covariance (CovMatrix): Covariance model to use.
            data (object): Data model; called to obtain vector errors given parameters.
            parameter_values_dict (dict): Parameter values for coefficient evaluation.
            fisher_properties (dict, optional): Options including `inversion_method`.
            covariance_prefactor_dict (dict, optional): Prefactors per block (gg/gv/vv).

        Returns:
            FisherMatrix: Ready-to-use Fisher matrix builder instance.
        """
        if covariance.matrix_form is False and covariance.emulator_flag is False:
            covariance.compute_matrix_covariance()
        if (
            covariance.compute_covariance_sum is None
            or covariance.compute_covariance_sum_jit is None
        ):
            covariance.init_compute_covariance_sum()

        fisher_properties = {
            **cls._default_fisher_properties,
            **fisher_properties,
        }

        _, vector_variance = data.give_data_and_variance(
            parameter_values_dict,
        )
        covariance_sum = covariance.compute_covariance_sum(
            parameter_values_dict,
            vector_variance,
            covariance_prefactor_dict=covariance_prefactor_dict,
        )

        inverse_covariance_sum = eval(
            f"inverse_covariance_{fisher_properties['inversion_method']}"
        )(covariance_sum)

        return cls(
            covariance=covariance,
            inverse_covariance_sum=inverse_covariance_sum,
            data_free_par=data.free_par,
            parameter_values_dict=parameter_values_dict,
        )

    def compute_covariance_derivative(
        self,
        partial_coefficients_dict_param,
    ):
        """Assemble covariance derivative matrix for a single parameter.

        Uses model-kind-specific blocks and partial derivative coefficients to form
        $\partial_i C$ for the parameter `i`.

        Args:
            partial_coefficients_dict_param (dict): Coefficients for blocks `gg/gv/vv`.

        Returns:
            numpy.ndarray: Covariance derivative matrix $\partial_i C$.

        Raises:
            ValueError: If the covariance model kind is unsupported.
        """

        if self.covariance.model_kind == "density":
            covariance_derivative_sum = jnp.sum(
                jnp.array(
                    [
                        partial_coefficients_dict_param["gg"][i] * cov
                        for i, cov in enumerate(self.covariance.covariance_dict["gg"])
                    ]
                ),
                axis=0,
            )

        elif self.covariance.model_kind == "velocity":
            covariance_derivative_sum = jnp.sum(
                jnp.array(
                    [
                        partial_coefficients_dict_param["vv"][i] * cov
                        for i, cov in enumerate(self.covariance.covariance_dict["vv"])
                    ]
                ),
                axis=0,
            )

        elif self.covariance.model_kind in ["density_velocity", "full"]:
            number_densities = self.covariance.number_densities
            number_velocities = self.covariance.number_velocities

            if self.covariance.model_kind == "density_velocity":
                covariance_derivative_sum_gv = jnp.zeros(
                    (number_densities, number_velocities)
                )
            elif self.covariance.model_kind == "full":
                covariance_derivative_sum_gv = jnp.sum(
                    [
                        partial_coefficients_dict_param["gv"][i] * cov
                        for i, cov in enumerate(self.covariance.covariance_dict["gv"])
                    ],
                    axis=0,
                )
            covariance_derivative_sum_gg = jnp.sum(
                [
                    partial_coefficients_dict_param["gg"][i] * cov
                    for i, cov in enumerate(self.covariance.covariance_dict["gg"])
                ],
                axis=0,
            )

            covariance_derivative_sum_vv = jnp.sum(
                [
                    partial_coefficients_dict_param["vv"][i] * cov
                    for i, cov in enumerate(self.covariance.covariance_dict["vv"])
                ],
                axis=0,
            )
            covariance_derivative_sum_vg = covariance_derivative_sum_gv.T

            covariance_derivative_sum = jnp.block(
                [
                    [covariance_derivative_sum_gg, covariance_derivative_sum_gv],
                    [covariance_derivative_sum_vg, covariance_derivative_sum_vv],
                ]
            )
        else:
            log.add("Wrong model type in the loaded covariance.")

        return covariance_derivative_sum

    def compute_fisher_matrix(self, covariance_prefactor_dict=None):
        """Compute the Fisher matrix using covariance derivatives.

        Returns:
            tuple[list[str], numpy.ndarray]: Parameter names list and Fisher matrix
            of shape `(n_params, n_params)`.
        """
        coefficients = importlib.import_module(
            f"flip.covariance.analytical.{self.covariance.model_name}.fisher_terms"
        )
        partial_coefficients_dict = coefficients.get_partial_derivative_coefficients(
            self.covariance.model_kind,
            self.parameter_values_dict,
            variant=self.covariance.variant,
            covariance_prefactor_dict=covariance_prefactor_dict,
        )
        parameter_name_list = []
        covariance_derivative_sum_list = []

        for (
            parameter_name,
            partial_coefficients_dict_param,
        ) in partial_coefficients_dict.items():
            parameter_name_list.append(parameter_name)
            covariance_derivative_sum_list.append(
                jnp.dot(
                    self.inverse_covariance_sum,
                    self.compute_covariance_derivative(
                        partial_coefficients_dict_param,
                    ),
                )
            )

        fisher_matrix_size = len(partial_coefficients_dict.keys())
        fisher_matrix = jnp.zeros((fisher_matrix_size, fisher_matrix_size))
        tri_i, tri_j = jnp.triu_indices_from(fisher_matrix)

        for i, j in zip(tri_i, tri_j):
            if jax_installed:
                fisher_matrix = fisher_matrix.at[i, j].set(
                    0.5
                    * jnp.trace(
                        covariance_derivative_sum_list[i]
                        @ covariance_derivative_sum_list[j]
                    )
                )
                fisher_matrix = fisher_matrix.at[j, i].set(fisher_matrix[i, j])
            else:
                fisher_matrix[i][j] = 0.5 * jnp.trace(
                    covariance_derivative_sum_list[i]
                    @ covariance_derivative_sum_list[j]
                )
                fisher_matrix[j, i] = fisher_matrix[i, j]

        return parameter_name_list, fisher_matrix



import numpy as np
from flip.covariance import covariance, fitter
from flip.data import load_data_test

from flip import data_vector


def test_e2e_density(debug_return=False):
    n = 50
    coordinates_density, density_data = load_data_test.load_density_data(subsample=n)

    density_data_vector = data_vector.Dens(density_data)

    power_spectrum_dict = load_data_test.load_power_spectrum_dict()

    covariance_object = covariance.CovMatrix.init_from_flip(
        "adamsblake17plane",
        "density",
        power_spectrum_dict,
        coordinates_density=coordinates_density,
        coordinates_velocity=None,
        size_batch=50_000,
        number_worker=1,
    )

    vals = {"bs8": 1.0, "fs8": 0.5, "sigv": 0.0}
    _, data_variance = density_data_vector.give_data_and_variance()
    covariance_matrix = covariance_object.compute_covariance_sum(vals, data_variance)

    like_props = {"inversion_method": "cholesky"}
    params = {"bs8": {"value": 1.0, "limit_low": 0.0, "fixed": False}}
    fit_minuit = fitter.FitMinuit.init_from_covariance(
        covariance_object,
        density_data_vector,
        params,
        likelihood_type="multivariate_gaussian",
        likelihood_properties=like_props,
    )
    vals1 = fit_minuit.run(migrad=True, hesse=False, minos=False, n_iter=1)
    vals2 = fit_minuit.run(migrad=True, hesse=False, minos=False, n_iter=1)
    if debug_return:
        return vals1

    reference_values = load_data_test.load_e2e_test_reference_values()["e2e_density"]

    assert covariance_matrix.shape == (n, n)
    np.testing.assert_allclose(covariance_matrix, covariance_matrix.T, atol=1e-12)
    assert np.all(np.diag(covariance_matrix) > 0)
    assert abs(vals1["bs8"] - vals2["bs8"]) < 0.05
    np.testing.assert_allclose(vals1["bs8"], reference_values["bs8"], rtol=5e-3, atol=0)



import numpy as np
import pytest
from flip.data import load_data_test

from flip import covariance, data_vector

model_to_test = [
    ("carreres23", "velocity"),
    ("adamsblake17plane", "density"),
    ("adamsblake17plane", "velocity"),
    ("adamsblake17plane", "full"),
    ("adamsblake20", "density"),
    ("adamsblake20", "velocity"),
    ("adamsblake20", "full"),
    ("ravouxcarreres", "velocity"),
    ("ravouxcarreres", "density"),
    ("ravouxcarreres", "full"),
    ("lai22", "density"),
    ("lai22", "velocity"),
    ("lai22", "full"),
]


def compute_covariance_metrics(
    model,
    model_type,
):

    if model_type == "density":
        coordinates_density, density_data = load_data_test.load_density_data(
            subsample=50
        )
        data_vector_obj = data_vector.Dens(density_data)
        coordinates_velocity = None

    elif model_type == "velocity":
        coordinates_velocity, velocity_data = load_data_test.load_velocity_data(
            subsample=50
        )
        data_vector_obj = data_vector.DirectVel(velocity_data)
        coordinates_density = None
    else:
        coordinates_velocity, velocity_data = load_data_test.load_velocity_data(
            subsample=50
        )
        coordinates_density, density_data = load_data_test.load_density_data(
            subsample=50
        )
        data_vector_obj = data_vector.DensVel(
            data_vector.Dens(density_data),
            data_vector.DirectVel(velocity_data),
        )

    power_spectrum_dict = load_data_test.load_power_spectrum_dict()

    additional_parameters = ()
    variant = None
    parameters = {"bs8": 1.0, "fs8": 0.5, "sigv": 0.0}
    if model in {"adamsblake20", "ravouxcarreres"}:
        parameters["beta_f"] = parameters["fs8"] / max(parameters["bs8"], 1e-6)
        additional_parameters = (15.0,)
        variant = "nobeta"
    if model == "lai22":
        parameters["sigg"] = 1.0
        parameters["beta_f"] = parameters["fs8"] / max(parameters["bs8"], 1e-6)
        variant = "nobeta"

    covariance_object = covariance.CovMatrix.init_from_flip(
        model,
        model_type,
        power_spectrum_dict,
        coordinates_density=coordinates_density,
        coordinates_velocity=coordinates_velocity,
        size_batch=50_000,
        number_worker=1,
        variant=variant,
        additional_parameters_values=additional_parameters,
    )

    _, data_variance = data_vector_obj.give_data_and_variance(
        parameters if "M_0" in data_vector_obj.free_par else {}
    )
    covariance_matrix = covariance_object.compute_covariance_sum(
        parameters, data_variance
    )
    return {
        "shape": [int(covariance_matrix.shape[0]), int(covariance_matrix.shape[1])],
        "trace": float(np.trace(covariance_matrix)),
        "diag_mean": float(np.mean(np.diag(covariance_matrix))),
        "entry_0_0": float(covariance_matrix[0, 0]),
        "entry_0_-1": float(covariance_matrix[0, -1]),
        "entry_mid_mid": float(
            covariance_matrix[
                covariance_matrix.shape[0] // 2, covariance_matrix.shape[1] // 2
            ]
        ),
    }


@pytest.mark.parametrize(
    "model,kind",
    model_to_test,
)
def test_covariance_reference_metrics(model, kind, debug_return=False):

    got = compute_covariance_metrics(model, kind)

    if debug_return:
        return got

    reference_values = load_data_test.load_covariance_test_reference_values()

    key = f"{model}:{kind}"
    assert key in reference_values, f"Missing reference metrics for {key}"
    ref = reference_values[key]
    assert got["shape"] == ref["shape"]
    np.testing.assert_allclose(got["trace"], ref["trace"], rtol=1e-6, atol=1e-8)
    np.testing.assert_allclose(got["diag_mean"], ref["diag_mean"], rtol=1e-6, atol=1e-8)
    np.testing.assert_allclose(got["entry_0_0"], ref["entry_0_0"], rtol=1e-6, atol=1e-8)
    np.testing.assert_allclose(
        got["entry_0_-1"], ref["entry_0_-1"], rtol=1e-6, atol=1e-8
    )
    np.testing.assert_allclose(
        got["entry_mid_mid"], ref["entry_mid_mid"], rtol=1e-6, atol=1e-8
    )



import numpy as np
from flip.data import load_data_test

from flip import covariance, data_vector


def test_covariance_assembly_density_velocity():

    coordinates_density, density_data = load_data_test.load_density_data(subsample=50)
    density_data_vector = data_vector.Dens(density_data)

    coordinates_velocity, velocity_data = load_data_test.load_velocity_data(
        subsample=50
    )
    velocity_data_vector = data_vector.DirectVel(velocity_data)

    density_velocity_data_vector = data_vector.DensVel(
        density_data_vector, velocity_data_vector
    )

    power_spectrum_dict = load_data_test.load_power_spectrum_dict()

    model_name = "ravouxcarreres"
    model_type = "density_velocity"
    sigmag_fiducial = 5.0

    covariance_object = covariance.CovMatrix.init_from_flip(
        model_name,
        model_type,
        power_spectrum_dict,
        coordinates_density=coordinates_density,
        coordinates_velocity=coordinates_velocity,
        size_batch=50_000,
        number_worker=1,
        additional_parameters_values=(sigmag_fiducial,),
        variant="nobeta",
    )

    # Assemble total covariance with simple coefficients
    coefficients = {"bs8": 1.0, "fs8": 1.0, "sigv": 0.0}
    _, data_variance = density_velocity_data_vector.give_data_and_variance()
    covariance_matrix = covariance_object.compute_covariance_sum(
        coefficients, data_variance
    )

    # Check shapes and that cross-block is exactly zero (no gv provided)
    n_density = len(density_data_vector.data["density"])
    n_velocity = len(velocity_data_vector.data["velocity"])
    covariance_density_density = covariance_matrix[:n_density, :n_density]
    covariance_velocity_velocity = covariance_matrix[n_density:, n_density:]
    covariance_density_velocity = covariance_matrix[:n_density, n_density:]

    np.testing.assert_allclose(covariance_density_velocity, 0.0, atol=0.0)
    assert covariance_matrix.shape == (n_density + n_velocity, n_density + n_velocity)
    assert np.all(np.diag(covariance_density_density) > 0)
    assert np.all(np.diag(covariance_velocity_velocity) > 0)


def test_covariance_assembly_full():

    coordinates_density, density_data = load_data_test.load_density_data(subsample=50)
    density_data_vector = data_vector.Dens(density_data)

    coordinates_velocity, velocity_data = load_data_test.load_velocity_data(
        subsample=50
    )
    velocity_data_vector = data_vector.DirectVel(velocity_data)

    density_velocity_data_vector = data_vector.DensVel(
        density_data_vector, velocity_data_vector
    )

    power_spectrum_dict = load_data_test.load_power_spectrum_dict()

    model_name = "ravouxcarreres"
    model_type = "full"
    sigmag_fiducial = 5.0

    covariance_object = covariance.CovMatrix.init_from_flip(
        model_name,
        model_type,
        power_spectrum_dict,
        coordinates_density=coordinates_density,
        coordinates_velocity=coordinates_velocity,
        size_batch=50_000,
        number_worker=1,
        additional_parameters_values=(sigmag_fiducial,),
        variant="nobeta",
    )

    coefficients = {"bs8": 1.0, "fs8": 1.0, "sigv": 0.0}
    _, data_variance = density_velocity_data_vector.give_data_and_variance()
    covariance_matrix = covariance_object.compute_covariance_sum(
        coefficients, data_variance
    )

    n_density = len(density_data_vector.data["density"])
    covariance_density_velocity = covariance_matrix[:n_density, n_density:]

    print(np.min(np.abs(covariance_density_velocity)))

    assert np.any(np.abs(covariance_density_velocity) > 0)



import numpy as np

from flip.utils import create_log

log = create_log()

try:
    import GPy

    gpy_installed = True
except ImportError:
    gpy_installed = False
    log.add(
        "Install GPy to use the gpmatrix emulator",
        level="warning",
    )


_emulator_type = "matrix"

if gpy_installed:
    default_kernel_variance = GPy.kern.Exponential(1) + GPy.kern.Poly(1)
    default_kernel_non_diagonal = GPy.kern.Exponential(1) + GPy.kern.Poly(1)
else:
    default_kernel_variance = None
    default_kernel_non_diagonal = None


def train(
    square_covariance,
    output_variance,
    output_non_diagonal,
    parameter_values,
    kernel_variance=None,
    kernel_non_diagonal=None,
    verbose=False,
    num_restarts_variance=None,
    num_restarts_non_diagonal=None,
    **kwargs,
):
    """Train Gaussian Process emulators for covariance matrices.

    Trains one GP per covariance term. For square covariances (gg or vv), it
    trains both a GP for the diagonal element (variance) and one for the
    non-diagonal elements. For rectangular covariances (gv), only the
    non-diagonal GP is trained.

    Args:
        square_covariance: Whether the covariance is square (gg or vv) or not (gv).
        output_variance: Array of shape ``(n_terms, n_samples)`` containing diagonal values per sample. Ignored if not square.
        output_non_diagonal: Array of shape ``(n_terms, n_samples, n_nd)`` containing flattened non-diagonal values per sample.
        parameter_values: Array of shape ``(n_samples, n_params)`` with emulator input parameters.
        kernel_variance: Optional GPy kernel to use for variance models. Defaults to ``Exponential + Poly`` if GPy is available.
        kernel_non_diagonal: Optional GPy kernel to use for non-diagonal models. Defaults to ``Exponential + Poly`` if GPy is available.
        verbose: If True, prints optimization messages.
        num_restarts_variance: Optional number of optimizer restarts for variance GPs.
        num_restarts_non_diagonal: Optional number of optimizer restarts for non-diagonal GPs.
        **kwargs: Extra keyword arguments (unused).

    Returns:
        Tuple containing:
        - ``gp_models_variance``: List of variance GP models or ``None``.
        - ``gp_models_non_diagonal``: List of non-diagonal GP models.
        - ``gp_evaluation_dictionary_variance``: Per-term evaluation dictionaries or ``None``.
        - ``gp_evaluation_dictionary_non_diagonal``: Per-term evaluation dictionaries.

    Raises:
        ImportError: If GPy is not installed when training is requested.
    """

    if square_covariance:
        if kernel_variance is None:
            kernel_variance = default_kernel_variance
        gp_models_variance = []
        gp_evaluation_dictionary_variance = [None for j in range(len(output_variance))]
    else:
        gp_models_variance = None
        gp_evaluation_dictionary_variance = None
    if kernel_non_diagonal is None:
        kernel_non_diagonal = default_kernel_non_diagonal

    gp_models_non_diagonal = []
    gp_evaluation_dictionary_non_diagonal = [
        None for j in range(len(output_non_diagonal))
    ]

    for j in range(len(output_non_diagonal)):

        if square_covariance:
            model_variance = GPy.models.GPRegression(
                parameter_values,
                output_variance[j][:, np.newaxis],
                kernel_variance,
            )
            model_variance.optimize(messages=verbose)
            if num_restarts_variance is not None:
                model_variance.optimize_restarts(num_restarts=num_restarts_variance)

            gp_models_variance.append(model_variance)

        model_non_diagonal = GPy.models.GPRegression(
            parameter_values, output_non_diagonal[j], kernel_non_diagonal
        )
        model_non_diagonal.optimize(messages=verbose)
        if num_restarts_non_diagonal is not None:
            model_non_diagonal.optimize_restarts(num_restarts=num_restarts_non_diagonal)

        gp_models_non_diagonal.append(model_non_diagonal)

    return (
        gp_models_variance,
        gp_models_non_diagonal,
        gp_evaluation_dictionary_variance,
        gp_evaluation_dictionary_non_diagonal,
    )


def evaluate(
    model,
    evaluation_value,
    evaluation_dictionary,
):
    """Evaluate a trained GP model at given parameter values.

    Args:
        model: Trained GPy GPRegression model.
        evaluation_value: Array of shape ``(1, n_params)`` with input parameters.
        evaluation_dictionary: Placeholder for API consistency (unused for GPy).

    Returns:
        Tuple ``(mean, std)`` as returned by ``GPy.models.GPRegression.predict``.
    """
    output = model.predict(evaluation_value)
    return output



import mpmath
import numpy
import scipy


def set_backend(module):
    global np, erf
    if module == "numpy":
        np = numpy
        erf = scipy.special.erf
    elif module == "mpmath":
        np = mpmath.mp
        erf = mpmath.erf


set_backend("numpy")


def M_gg_0_0_0():
    def func(k):
        return 1

    return func


def N_gg_0_0_0(theta, phi):
    return 1


def M_gv_0_1_0():
    def func(k):
        return (100 / 3) / k

    return func


def N_gv_0_1_0(theta, phi):
    return -3 * np.cos(phi + (1 / 2) * theta)


def M_vv_0_0_0():
    def func(k):
        return (10000 / 9) / k**2

    return func


def N_vv_0_0_0(theta, phi):
    return 3 * np.cos(theta)


def M_vv_0_2_0():
    def func(k):
        return (10000 / 9) / k**2

    return func


def N_vv_0_2_0(theta, phi):
    return (9 / 2) * np.cos(2 * phi) + (3 / 2) * np.cos(theta)


dictionary_terms = {"gg": ["0"], "gv": ["0"], "vv": ["0"]}
dictionary_lmax = {"gg": [2], "gv": [2], "vv": [2]}
dictionary_subterms = {
    "gg_0_0": 1,
    "gg_0_1": 0,
    "gg_0_2": 0,
    "gv_0_0": 0,
    "gv_0_1": 1,
    "gv_0_2": 0,
    "vv_0_0": 1,
    "vv_0_1": 0,
    "vv_0_2": 1,
}
multi_index_model = False
regularize_M_terms = None



import numpy as np
from flip.covariance.analytical.lai22.flip_terms import dictionary_terms


def get_coefficients(
    parameter_values_dict,
    model_kind,
    variant=None,
    covariance_prefactor_dict=None,
):
    coefficients_dict = {}
    if model_kind in ["density", "full", "density_velocity"]:
        coefficients_dict["gg"] = []
        gg_terms = dictionary_terms["gg"]
        sigg = parameter_values_dict["sigg"]
        for gg_term in gg_terms:
            term_index, m_index = np.array(gg_term.split("_")).astype(int)
            if term_index == 0:
                coefficients_dict["gg"].append(
                    parameter_values_dict["bs8"] ** 2 * sigg ** (2 * m_index)
                )
            elif term_index == 1:
                if variant == "nobeta":
                    coefficients_dict["gg"].append(
                        parameter_values_dict["bs8"]
                        * parameter_values_dict["fs8"]
                        * sigg ** (2 * m_index)
                    )
                else:
                    coefficients_dict["gg"].append(
                        parameter_values_dict["bs8"] ** 2
                        * parameter_values_dict["beta_f"]
                        * sigg ** (2 * m_index)
                    )
            elif term_index == 2:
                if variant == "nobeta":
                    coefficients_dict["gg"].append(
                        parameter_values_dict["fs8"] ** 2 * sigg ** (2 * m_index)
                    )
                else:
                    coefficients_dict["gg"].append(
                        parameter_values_dict["bs8"] ** 2
                        * parameter_values_dict["beta_f"] ** 2
                        * sigg ** (2 * m_index)
                    )
    if model_kind in ["full"]:
        coefficients_dict["gv"] = []
        gv_terms = dictionary_terms["gv"]
        sigg = parameter_values_dict["sigg"]
        for gv_term in gv_terms:
            term_index, m_index = np.array(gv_term.split("_")).astype(int)
            if term_index == 0:
                coefficients_dict["gv"].append(
                    parameter_values_dict["bs8"]
                    * parameter_values_dict["fs8"]
                    * sigg ** (2 * m_index)
                )
            elif term_index == 1:
                if variant == "nobeta":
                    coefficients_dict["gv"].append(
                        parameter_values_dict["fs8"] ** 2 * sigg ** (2 * m_index)
                    )
                else:
                    coefficients_dict["gv"].append(
                        parameter_values_dict["bs8"]
                        * parameter_values_dict["fs8"]
                        * parameter_values_dict["beta_f"]
                        * sigg ** (2 * m_index)
                    )

    if model_kind in ["velocity", "full", "density_velocity"]:
        coefficients_dict["vv"] = [parameter_values_dict["fs8"] ** 2]
    return coefficients_dict


def get_diagonal_coefficients(parameter_values_dict, model_kind):
    coefficients_dict = {}
    if model_kind in ["density", "full", "density_velocity"]:
        coefficients_dict["gg"] = 0.0
    if model_kind in ["velocity", "full", "density_velocity"]:
        coefficients_dict["vv"] = parameter_values_dict["sigv"] ** 2
    return coefficients_dict



import importlib

import numpy as np

from flip.covariance import cov_utils
from flip.utils import create_log

log = create_log()

_avail_emulator_models = [
    "gpmatrix",
    "nnmatrix",
    "skgpmatrix",
]


def compute_cov(
    emulator_model_name,
    covariance_list,
    parameter_values,
    emulator_parameter_names,
    covariance_type,
    test_emulator=False,
    **kwargs,
):
    """Train an emulator for a given covariance type and return evaluators.

    Args:
        emulator_model_name: Backend name, one of ``gpmatrix``, ``nnmatrix``, ``skgpmatrix``.
        covariance_list: List of covariance objects with precomputed matrices.
        parameter_values: Array ``(n_samples, n_params)`` training inputs.
        emulator_parameter_names: Names of parameters to pass at evaluation time.
        covariance_type: One of ``"gg"``, ``"vv"``, ``"gv"`` indicating block type.
        test_emulator: If True, runs a quick evaluation vs. training data and logs errors.
        **kwargs: Extra keyword arguments forwarded to backend ``train``.

    Returns:
        List of callables, one per covariance term, taking a parameter dict and returning a matrix.

    Raises:
        ValueError: If list lengths mismatch or model name is invalid.
    """
    if len(covariance_list) != len(parameter_values):
        raise ValueError(
            "covariance_list and parameter_values must have the same length."
        )
    if emulator_model_name not in _avail_emulator_models:
        log.add(
            f"Model {emulator_model_name} not available."
            f"Please choose between: {_avail_emulator_models}"
        )

    emulator_module = importlib.import_module(
        f"flip.covariance.emulators.{emulator_model_name}"
    )

    # emulator_type = emulator_module._emulator_type
    # Only implemented for matrix emulators, so far.

    square_covariance, emulator_output_variance, emulator_output_non_diagonal = (
        prepare_covariance_matrices(
            covariance_list,
            covariance_type,
        )
    )

    models = emulator_module.train(
        square_covariance,
        emulator_output_variance,
        emulator_output_non_diagonal,
        parameter_values,
        **kwargs,
    )

    evaluation_functions = return_evaluation_functions(
        models,
        emulator_module,
        emulator_parameter_names,
        covariance_type,
        **kwargs,
    )

    if test_emulator:
        parameter_values_dict = [
            {
                emulator_parameter_names[j]: parameter_values[i][j]
                for j in range(len(emulator_parameter_names))
            }
            for i in range(len(parameter_values))
        ]
        test_training(
            evaluation_functions,
            covariance_list,
            parameter_values_dict,
            covariance_type,
        )

    return evaluation_functions


def prepare_covariance_matrices(
    covariance_list,
    covariance_type,
):
    """Extract and stack covariance training targets for emulator backends.

    Converts each covariance object into arrays for diagonal and non-diagonal
    parts depending on the covariance type.

    Args:
        covariance_list: List of covariance objects.
        covariance_type: Block type ``("gg"|"vv"|"gv")``.

    Returns:
        Tuple ``(square_covariance, output_variance, output_non_diagonal)`` where
        arrays have shapes compatible with backend trainers.
    """
    if covariance_type[0] == covariance_type[1]:
        square_covariance = True
    else:
        square_covariance = False

    emulator_output_non_diagonal = [
        [] for _ in range(len(covariance_list[0].covariance_dict[covariance_type]))
    ]
    if square_covariance:
        emulator_output_variance = [
            [] for _ in range(len(covariance_list[0].covariance_dict[covariance_type]))
        ]
    else:
        emulator_output_variance = None

    for i in range(len(covariance_list)):
        covariance_list[i].compute_flat_covariance(verbose=False)
        for j in range(len(covariance_list[i].covariance_dict[covariance_type])):

            if square_covariance:
                emulator_output_variance[j].append(
                    covariance_list[i].covariance_dict[covariance_type][j][0]
                )

                emulator_output_non_diagonal[j].append(
                    covariance_list[i].covariance_dict[covariance_type][j][1:]
                )
            else:
                emulator_output_non_diagonal[j].append(
                    covariance_list[i].covariance_dict[covariance_type][j]
                )

    return (
        square_covariance,
        np.array(emulator_output_variance),
        np.array(emulator_output_non_diagonal),
    )


def return_evaluation_functions(
    models,
    emulator_module,
    emulator_parameter_names,
    covariance_type,
    **kwargs,
):
    """Wrap backend models into per-term evaluation functions.

    Each returned function accepts a parameter dict mapping names to values and
    returns either a square matrix (gg/vv) or a flattened vector (gv).

    Args:
        models: Tuple of backend-trained models and evaluation dictionaries.
        emulator_module: Module providing ``evaluate(model, x, dict)`` function.
        emulator_parameter_names: Names used to extract parameter order from dict.
        covariance_type: Block type ``("gg"|"vv"|"gv")``.
        **kwargs: Extra keyword arguments (unused).

    Returns:
        List of evaluation callables for each covariance term.
    """

    if covariance_type[0] == covariance_type[1]:
        square_covariance = True
        models_variance, models_non_diagonal = models[0], models[1]
        evaluation_dictionary_variance, evaluation_dictionary_non_diagonal = (
            models[2],
            models[3],
        )
    else:
        square_covariance = False
        models_non_diagonal = models[1]
        evaluation_dictionary_non_diagonal = models[3]

    evaluation_functions = []

    for j in range(len(models_non_diagonal)):
        if square_covariance:
            model_variance = models_variance[j]
        model_non_diagonal = models_non_diagonal[j]

        if square_covariance:

            def evaluation_function(
                parameter_values_dict,
                return_emulator_variance=False,
            ):

                parameter_value = [
                    parameter_values_dict[name] for name in emulator_parameter_names
                ]
                evaluation_value = np.array(parameter_value)[np.newaxis, :]
                output_non_diagonal = emulator_module.evaluate(
                    model_non_diagonal,
                    evaluation_value,
                    evaluation_dictionary_non_diagonal[j],
                )
                output_covariance = output_non_diagonal[0][0]
                output_diagonal = emulator_module.evaluate(
                    model_variance,
                    evaluation_value,
                    evaluation_dictionary_variance[j],
                )
                output_covariance = np.insert(
                    output_covariance, 0, output_diagonal[0][0]
                )
                output_covariance = cov_utils.return_matrix_covariance(
                    output_covariance
                )
                if return_emulator_variance:
                    variance_emulator = [
                        output_diagonal[1][0, 0],
                        output_non_diagonal[1][0, 0],
                    ]
                    return output_covariance, variance_emulator
                else:
                    return output_covariance

        else:

            def evaluation_function(
                parameter_value,
                return_emulator_variance=False,
            ):
                evaluation_value = np.array(parameter_value)[np.newaxis, np.newaxis]
                output_non_diagonal = emulator_module.evaluate(
                    model_non_diagonal,
                    evaluation_value,
                    evaluation_dictionary_non_diagonal[j],
                )
                output_covariance = output_non_diagonal[0][0]
                variance_emulator = output_non_diagonal[1][0, 0]

                if return_emulator_variance:
                    return output_covariance, variance_emulator
                else:
                    return output_covariance

        evaluation_functions.append(evaluation_function)

    return evaluation_functions


def test_training(
    evaluation_functions,
    covariance_list,
    parameter_values_dict,
    covariance_type,
    verbose=True,
    **kwargs,
):
    """Compute emulator errors against training covariances for sanity checks.

    Args:
        evaluation_functions: List of term evaluators returned by ``return_evaluation_functions``.
        covariance_list: List of covariance objects (ground truth).
        parameter_values_dict: List of parameter dicts matching training inputs.
        covariance_type: Block type ``("gg"|"vv"|"gv")``.
        verbose: If True, logs per-term maximal relative errors.
        **kwargs: Extra keyword arguments (unused).

    Returns:
        Nested list of errors per sample and term: ``[ [err_diag, rel_err_diag, err_all, rel_err_all], ... ]``.
    """
    if covariance_type[0] == covariance_type[1]:
        square_covariance = True
    else:
        square_covariance = False
    errors = []
    for i in range(len(parameter_values_dict)):
        covariance_list[i].compute_matrix_covariance(verbose=False)
        errors.append([])
        for j in range(len(evaluation_functions)):
            evaluation = evaluation_functions[j](parameter_values_dict[i])
            covariance = covariance_list[i].covariance_dict[covariance_type][j]

            if square_covariance:
                error_diagonal = np.abs(evaluation[0][0] - covariance[0][0])
                relative_error_diagonal = np.abs(
                    evaluation[0][0] - covariance[0][0]
                ) / np.abs(covariance[0][0])
            else:
                error_diagonal = None
                relative_error_diagonal = None
            error_all = np.max(np.abs(evaluation - covariance))
            relative_error_all = np.max(
                np.abs(evaluation - covariance) / np.max(np.abs(covariance))
            )
            if verbose:
                log.add(
                    (
                        f"For parameter {i} and covariance type {covariance_type}\n"
                        f"Covariance term {j} gives: \n"
                        f"Maximal relative error all {relative_error_all}\n"
                    ),
                    level="info",
                )
                if square_covariance:
                    log.add(
                        f"Maximal relative error diagonal {relative_error_diagonal}, \n",
                        level="info",
                    )

            errors[i].append(
                [
                    error_diagonal,
                    relative_error_diagonal,
                    error_all,
                    relative_error_all,
                ]
            )

    return errors


def generate_covariance(
    emulator_model_name,
    model_kind,
    covariance_list,
    parameter_values,
    emulator_parameter_names,
    **kwargs,
):
    """Generate emulator evaluators for requested covariance blocks.

    Args:
        emulator_model_name: Backend name.
        model_kind: One of ``"density"``, ``"velocity"``, ``"full"``, ``"density_velocity"``.
        covariance_list: List of covariance objects for training.
        parameter_values: Array ``(n_samples, n_params)`` training inputs.
        emulator_parameter_names: Parameter names used at evaluation.
        **kwargs: Forwarded to backend trainers.

    Returns:
        Dict with keys among ``{"gg","vv","gv"}`` mapping to lists of evaluator functions.
    """

    emulator_covariance_dict = {}

    if model_kind in ["density", "full", "density_velocity"]:
        emulator_covariance_dict["gg"] = compute_cov(
            emulator_model_name,
            covariance_list,
            parameter_values,
            emulator_parameter_names,
            "gg",
            **kwargs,
        )

    if model_kind in ["velocity", "full", "density_velocity"]:
        emulator_covariance_dict["vv"] = compute_cov(
            emulator_model_name,
            covariance_list,
            parameter_values,
            emulator_parameter_names,
            "vv",
            **kwargs,
        )

    if model_kind == "full":
        emulator_covariance_dict["gv"] = compute_cov(
            emulator_model_name,
            covariance_list,
            parameter_values,
            emulator_parameter_names,
            "gv",
            **kwargs,
        )

    return emulator_covariance_dict



import numpy as np

from flip.utils import create_log

log = create_log()

try:
    from cosmoprimo import Fourier
except ImportError:
    log.add(
        "Install cosmoprimo https://github.com/cosmodesi/cosmoprimo to use cosmoprimo_engine.py module",
        level="warning",
    )


def get_fiducial_fs8(cosmology, redshift):
    """Return fiducial $f\sigma_8$ using Cosmoprimo at redshift.

    Args:
        cosmology: Cosmoprimo cosmology instance.
        redshift (float): Target redshift.

    Returns:
        float: $\sigma_8(z) \times f(z)$.
    """
    return cosmology.sigma8_z(redshift) * cosmology.growth_rate(redshift)


def get_fiducial_s8(cosmology, redshift):
    """Return fiducial $\sigma_8$ using Cosmoprimo at redshift.

    Args:
        cosmology: Cosmoprimo cosmology instance.
        redshift (float): Target redshift.

    Returns:
        float: $\sigma_8(z)$.
    """
    return cosmology.sigma8_z(redshift)


def compute_power_spectrum(
    cosmology,
    redshift,
    minimal_wavenumber,
    maximal_wavenumber,
    number_points,
    non_linear_model=None,
    logspace=True,
):
    """Compute linear/non-linear $P(k)$ using Cosmoprimo Fourier interface.

    Args:
        cosmology: Cosmoprimo cosmology instance.
        redshift (float): Redshift for $P(k)$.
        minimal_wavenumber (float): Minimum $k$ in h/Mpc.
        maximal_wavenumber (float): Maximum $k$ in h/Mpc.
        number_points (int): Number of $k$ samples.
        non_linear_model (str|None): If not None, compute non-linear spectrum.
        logspace (bool): Sample $k$ in log-space when True.

    Returns:
        tuple: `(k, P_lin, P_nl_or_None, fiducial_dict)`.

    Raises:
        ValueError: If a dict is provided instead of a Cosmology instance.
    """
    if type(cosmology) is dict:
        raise ValueError(
            "power_spectrum_settings should be an instance of Cosmology, not a dict"
        )
    if logspace:
        wavenumber = np.logspace(
            np.log10(minimal_wavenumber),
            np.log10(maximal_wavenumber),
            number_points,
        )
    else:
        wavenumber = np.linspace(
            minimal_wavenumber,
            maximal_wavenumber,
            number_points,
        )

    fourier = Fourier(cosmology, engine="class")
    power_spectrum_linear = fourier.pk_interpolator(of=("delta_m"))(
        wavenumber, z=redshift
    )
    if non_linear_model is not None:
        power_spectrum_non_linear = fourier.pk_interpolator(
            of=("delta_m"), non_linear=True
        )(wavenumber, z=redshift)
    else:
        power_spectrum_non_linear = None

    fs8_fiducial = get_fiducial_fs8(cosmology, redshift)
    s8_fiducial = get_fiducial_s8(cosmology, redshift)
    fiducial = {"fsigma_8": fs8_fiducial, "sigma_8": s8_fiducial}

    if non_linear_model is None:
        power_spectrum_non_linear = None

    return (
        wavenumber,
        power_spectrum_linear,
        power_spectrum_non_linear,
        fiducial,
    )
